import nengo
import numpy as np
import pytest

import nengo_extras.keras
from nengo_extras.keras import SequentialNetwork, kmodel_compute_shapes


@pytest.mark.parametrize("noise_model", ("none", "alpharc"))
def test_softlif_layer(noise_model, plt, allclose):
    keras = pytest.importorskip("keras")

    nx = 256
    ni = 500

    params = dict(sigma=0.02, amplitude=0.063, tau_rc=0.05, tau_ref=0.001)
    layer_params = dict(params)
    layer_params["noise_model"] = noise_model
    if noise_model == "alpharc":
        layer_params["tau_s"] = 0.001

    model = keras.models.Sequential()
    model.add(nengo_extras.keras.SoftLIF(input_shape=(nx,), **layer_params))

    x = np.linspace(-10, 30, nx)
    y = model.predict(np.ones((ni, 1)) * x)
    y0 = nengo_extras.neurons.SoftLIFRate(**params).rates(x, 1.0, 1.0).ravel()

    plt.plot(x, y.mean(axis=0), "b")
    if noise_model != "none":
        plt.plot(x, np.percentile(y, 5, axis=0), "b:")
        plt.plot(x, np.percentile(y, 95, axis=0), "b:")
    plt.plot(x, y0, "k--")

    assert np.isfinite(y).all()
    if noise_model == "none":
        assert allclose(y, y0, atol=1e-3, rtol=1e-3)
    else:
        assert allclose(y.mean(axis=0), y0, atol=1e-1, rtol=1e-2)


@pytest.mark.parametrize("noise_model", ("none", "alpharc"))
def test_softlif_derivative(noise_model, plt, allclose):
    tf = pytest.importorskip("tensorflow")
    keras = pytest.importorskip("keras")
    K = keras.backend

    # TODO: make this work with eager execution
    tf.compat.v1.disable_eager_execution()

    params = dict(sigma=0.02, amplitude=0.063, tau_rc=0.05, tau_ref=0.001)
    layer_params = dict(params)
    layer_params["noise_model"] = noise_model

    model = keras.models.Sequential()
    model.add(nengo_extras.keras.SoftLIF(input_shape=(1,), **layer_params))

    x = model.layers[0].input
    y = model.layers[0].output
    df = K.function([x], K.gradients(K.sum(y), [x]))

    x = np.linspace(-10, 30, 1024).reshape((-1, 1))
    dy = df([x])[0]
    dy0 = nengo_extras.neurons.SoftLIFRate(**params).derivative(x, 1.0, 1.0)

    plt.plot(x, dy)
    plt.plot(x, dy0, "k--")

    assert np.isfinite(dy).all()
    assert allclose(dy, dy0, atol=1e-2, rtol=1e-3)


def test_conv2d_layer(Simulator, seed, rng):
    keras = pytest.importorskip("keras")

    np.random.seed(seed)  # for Keras weights

    data_format = "channels_first"
    img_shape = (1, 28, 28)
    presentation_time = 0.01

    nb_filters = 4
    nb_conv = 3

    kmodel = keras.models.Sequential()
    kmodel.add(
        keras.layers.Conv2D(
            nb_filters,
            (nb_conv, nb_conv),
            padding="valid",
            strides=(2, 2),
            input_shape=img_shape,
            data_format=data_format,
        )
    )

    X = rng.uniform(-1, 1, size=(1,) + img_shape)
    output_shape = kmodel_compute_shapes(kmodel, X.shape)[-1][1:]

    with nengo.Network() as model:
        u = nengo.Node(nengo.processes.PresentInput(X, presentation_time))
        knet = SequentialNetwork(kmodel, synapse=None, lif_type="lifrate")
        nengo.Connection(u, knet.input, synapse=None)
        output_p = nengo.Probe(knet.output)

    with Simulator(model) as sim:
        sim.run(presentation_time)

    y0 = kmodel.predict(X)[0]
    y1 = sim.data[output_p][-1].reshape(output_shape)
    assert np.allclose(y1, y0, atol=1e-5, rtol=1e-5)



import logging
import warnings

import numpy as np

from nengo_extras import reqs

if reqs.HAS_NENGO_DL:
    import tensorflow as tf
    from nengo_dl.neuron_builders import LIFBuilder, SpikingRectifiedLinearBuilder

else:

    class LIFBuilder:
        pass

    class SpikingRectifiedLinearBuilder:
        pass


logger = logging.getLogger(__name__)


class NeuronOutputNoise:
    """Noise added to the output of a rate neuron.

    Often used when training deep networks with rate neurons for final
    implementation in spiking neurons to simulate the variability
    caused by spiking.
    """


class LowpassRCNoise(NeuronOutputNoise):
    """Noise model combining Lowpass synapse and neuron membrane filters.

    Samples "noise" (i.e. variability) from a regular spike train filtered
    by the following transfer function, where :math:`\\tau_{rc}` is the
    membrane time constant and :math:`\\tau_s` is the synapse time constant:

    .. math::

        H(s) = [(\\tau_s s + 1) (\\tau_{rc} s + 1)]^{-1}

    See [1]_ for background and derivations.

    Attributes
    ----------
    tau_s : float
        Time constant for Lowpass synaptic filter.

    References
    ----------
    .. [1] E. Hunsberger (2018) "Spiking Deep Neural Networks: Engineered and
       Biological Approaches to Object Recognition." PhD thesis. pp. 106--113.
       (http://compneuro.uwaterloo.ca/publications/hunsberger2018.html)
    """

    def __init__(self, tau_s):
        self.tau_s = tau_s

    def __repr__(self):
        return "%s(tau_s=%s)" % (type(self).__name__, self.tau_s)


class AlphaRCNoise(NeuronOutputNoise):
    """Noise model combining Alpha synapse and neuron membrane filters.

    Samples "noise" (i.e. variability) from a regular spike train filtered
    by the following transfer function, where :math:`\\tau_{rc}` is the
    membrane time constant and :math:`\\tau_s` is the synapse time constant:

    .. math::

        H(s) = [(\\tau_s s + 1)^2 (\\tau_{rc} s + 1)]^{-1}

    See [1]_ for background and derivations.

    Attributes
    ----------
    tau_s : float
        Time constant for Alpha synaptic filter.

    References
    ----------
    .. [1] E. Hunsberger (2018) "Spiking Deep Neural Networks: Engineered and
       Biological Approaches to Object Recognition." PhD thesis. pp. 106--113.
       (http://compneuro.uwaterloo.ca/publications/hunsberger2018.html)
    """

    def __init__(self, tau_s):
        self.tau_s = tau_s

    def __repr__(self):
        return "%s(tau_s=%s)" % (type(self).__name__, self.tau_s)


class NoiseBuilder:
    """Build noise classes in ``nengo_dl``.

    Attributes
    ----------
    noise_models : list of NeuronOutputNoise
        The noise models used for each op/signal.
    """

    builders = {}

    def __init__(self, ops, noise_models):
        self.ops = ops
        self.noise_models = noise_models

    @classmethod
    def build(cls, ops):
        """Create a NoiseBuilder for the provided ops."""

        noise_models = [getattr(op.neurons, "nengo_dl_noise", None) for op in ops]
        model_type = type(noise_models[0]) if len(noise_models) > 0 else None
        if not all(type(m) is model_type for m in noise_models):
            raise NotImplementedError(
                "Multiple noise models for the same neuron type is not supported"
            )

        return cls.builders[model_type](ops, noise_models)

    @classmethod
    def register(cls, noise_builder):
        """A decorator for adding a class to the list of builders.

        Raises a warning if a builder already exists for the class.

        Parameters
        ----------
        noise_builder : NoiseBuilder
            The NoiseBuilder subclass that the decorated class builds.
        """

        def register_builder(build_cls):
            if noise_builder in cls.builders:
                warnings.warn(
                    "Type '%s' already has a builder. Overwriting." % noise_builder
                )
            cls.builders[noise_builder] = build_cls
            return build_cls

        return register_builder

    def build_pre(self, signals, config):
        self.dtype = signals.dtype
        self.np_dtype = self.dtype.as_numpy_dtype()

    def build_step(self, period, tau_rc=None):
        """Generate TensorFlow code to implement these noise models.

        Parameters
        ----------
        period : tf.Tensor
            The inter-spike periods of the neurons to add noise to.
        tau_rc : tf.Tensor
            The membrane time constant of the neurons (used by some noise
            models).
        """
        raise NotImplementedError("Subclass must implement")


@NoiseBuilder.register(type(None))
class NoNoiseBuilder(NoiseBuilder):
    """nengo_dl builder for if there is no noise model."""

    def build_step(self, period, tau_rc=None):
        return tf.math.reciprocal(period)


@NoiseBuilder.register(LowpassRCNoise)
class LowpassRCNoiseBuilder(NoiseBuilder):
    """nengo_dl builder for the LowpassRCNoise model."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        # tau_s is the time constant of the synaptic filter
        tau_s = np.concatenate(
            [
                model.tau_s * np.ones((1, op.J.shape[0]), dtype=self.np_dtype)
                for model, op in zip(self.noise_models, self.ops)
            ]
        )
        self.tau_s = tf.constant(tau_s, dtype=self.dtype)

    def build_step(self, period, tau_rc=None):
        d = tau_rc - self.tau_s
        u01 = tf.random.uniform(tf.shape(period))
        t = u01 * period
        q_rc = tf.exp(-t / tau_rc)
        q_s = tf.exp(-t / self.tau_s)
        r_rc1 = -tf.math.expm1(-period / tau_rc)  # 1 - exp(-period/tau_rc)
        r_s1 = -tf.math.expm1(-period / self.tau_s)  # 1 - exp(-period/tau_s)
        return (1.0 / d) * (q_rc / r_rc1 - q_s / r_s1)


@NoiseBuilder.register(AlphaRCNoise)
class AlphaRCNoiseBuilder(NoiseBuilder):
    """nengo_dl builder for the AlphaRCNoise model."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        # tau_s is the time constant of the synaptic filter
        tau_s = np.concatenate(
            [
                model.tau_s * np.ones((1, op.J.shape[0]), dtype=self.np_dtype)
                for model, op in zip(self.noise_models, self.ops)
            ]
        )
        self.tau_s = tf.constant(tau_s, dtype=self.dtype)

    def build_step(self, period, tau_rc=None):
        d = tau_rc - self.tau_s
        u01 = tf.random.uniform(tf.shape(period))
        t = u01 * period
        q_rc = tf.exp(-t / tau_rc)
        q_s = tf.exp(-t / self.tau_s)
        r_rc1 = -tf.math.expm1(-period / tau_rc)  # 1 - exp(-period/tau_rc)
        r_s1 = -tf.math.expm1(-period / self.tau_s)  # 1 - exp(-period/tau_s)

        pt = tf.where(
            period < 100 * self.tau_s, (period - t) * (1 - r_s1), tf.zeros_like(period)
        )
        qt = tf.where(t < 100 * self.tau_s, q_s * (t + pt), tf.zeros_like(t))
        rt = qt / (self.tau_s * d * r_s1 ** 2)
        rn = tau_rc * (q_rc / (d * d * r_rc1) - q_s / (d * d * r_s1)) - rt
        return rn


class LoihiLIFBuilder(LIFBuilder):
    """nengo_dl builder for the LoihiLIF neuron type.

    Attributes
    ----------
    spike_noise : NoiseBuilder
        Generator for any output noise associated with these neurons.
    """

    def __init__(self, ops):
        from nengo_loihi.neurons import (  # pylint: disable=import-outside-toplevel
            discretize_tau_rc,
            discretize_tau_ref,
        )

        self.discretize_tau_rc = discretize_tau_rc
        self.discretize_tau_ref = discretize_tau_ref

        super().__init__(ops)
        self.spike_noise = NoiseBuilder.build(ops)

    def build_pre(self, signals, config):
        super().build_pre(signals, config)
        self.spike_noise.build_pre(signals, config)

    def training_step(self, J, dt, **state):
        tau_ref = self.discretize_tau_ref(self.tau_ref, dt)
        tau_rc = self.discretize_tau_rc(self.tau_rc, dt)
        # Since LoihiLIF takes `ceil(period/dt)` the firing rate is
        # always below the LIF rate. Using `tau_ref1` in LIF curve makes
        # it the average of the LoihiLIF curve (rather than upper bound).
        tau_ref1 = tau_ref + 0.5 * dt

        J -= self.one

        # --- compute Loihi rates (for forward pass)
        period = tau_ref + tau_rc * tf.math.log1p(
            tf.math.reciprocal(tf.maximum(J, self.epsilon))
        )
        period = dt * tf.math.ceil(period / dt)
        loihi_rates = self.spike_noise.build_step(period, tau_rc=tau_rc)
        loihi_rates = tf.where(J > self.zero, loihi_rates, self.zeros)
        if self.amplitude is not None:
            loihi_rates *= self.amplitude

        # --- compute LIF rates (for backward pass)
        amplitude = self.one if self.amplitude is None else self.amplitude
        if self.config.lif_smoothing:
            js = J / self.sigma
            j_valid = js > -20
            js_safe = tf.where(j_valid, js, self.zeros)

            # softplus(js) = log(1 + e^js)
            z = tf.nn.softplus(js_safe) * self.sigma

            # as z->0
            #   z = s*log(1 + e^js) = s*e^js
            #   log(1 + 1/z) = log(1/z) = -log(s*e^js) = -js - log(s)
            q = tf.where(
                j_valid,
                tf.math.log1p(tf.math.reciprocal(z)),
                -js - tf.math.log(self.sigma),
            )

            rates = amplitude / (tau_ref1 + tau_rc * q)
        else:  # pragma: no cover
            rates = amplitude / (
                tau_ref1
                + tau_rc
                * tf.math.log1p(tf.math.reciprocal(tf.maximum(J, self.epsilon)))
            )
            rates = tf.where(J > self.zero, rates, self.zeros)

        # rates + stop_gradient(loihi_rates - rates) =
        #     loihi_rates on forward pass, rates on backwards
        return rates + tf.stop_gradient(loihi_rates - rates)

    def step(self, J, dt, voltage, refractory_time):
        tau_ref = self.discretize_tau_ref(self.tau_ref, dt)
        tau_rc = self.discretize_tau_rc(self.tau_rc, dt)

        delta_t = tf.clip_by_value(dt - refractory_time, self.zero, dt)
        voltage -= (J - voltage) * tf.math.expm1(-delta_t / tau_rc)

        spiked = voltage > self.one
        spikes = tf.cast(spiked, J.dtype) * self.alpha

        refractory_time = tf.where(spiked, tau_ref + self.zeros, refractory_time - dt)
        voltage = tf.where(spiked, self.zeros, tf.maximum(voltage, self.min_voltage))

        # we use stop_gradient to avoid propagating any nans (those get
        # propagated through the cond even if the spiking version isn't
        # being used at all)
        return (
            tf.stop_gradient(spikes),
            tf.stop_gradient(voltage),
            tf.stop_gradient(refractory_time),
        )


class LoihiSpikingRectifiedLinearBuilder(SpikingRectifiedLinearBuilder):
    """nengo_dl builder for the LoihiSpikingRectifiedLinear neuron type."""

    def build_pre(self, signals, config):
        super().build_pre(signals, config)

        self.zeros = tf.zeros(
            (signals.minibatch_size,) + self.J_data.shape, signals.dtype
        )

        self.epsilon = tf.constant(1e-15, dtype=signals.dtype)

        # copy these so that they're easily accessible in _step functions
        self.zero = signals.zero
        self.one = signals.one

    def training_step(self, J, dt, **state):
        # Since LoihiLIF takes `ceil(period/dt)` the firing rate is
        # always below the LIF rate. Using `tau_ref1` in LIF curve makes
        # it the average of the LoihiLIF curve (rather than upper bound).
        tau_ref1 = 0.5 * dt

        # --- compute Loihi rates (for forward pass)
        period = tf.math.reciprocal(tf.maximum(J, self.epsilon))
        loihi_rates = self.alpha / tf.math.ceil(period / dt)
        loihi_rates = tf.where(J > self.zero, loihi_rates, self.zeros)

        # --- compute RectifiedLinear rates (for backward pass)
        rates = (self.one if self.amplitude is None else self.amplitude) / (
            tau_ref1 + tf.math.reciprocal(tf.maximum(J, self.epsilon))
        )
        rates = tf.where(J > self.zero, rates, self.zeros)

        # rates + stop_gradient(loihi_rates - rates) =
        #     loihi_rates on forward pass, rates on backwards
        return rates + tf.stop_gradient(loihi_rates - rates)

    def step(self, J, dt, voltage):
        voltage += J * dt
        spiked = voltage > self.one
        spikes = tf.cast(spiked, J.dtype) * self.alpha
        voltage = tf.where(spiked, self.zeros, tf.maximum(voltage, 0))

        # we use stop_gradient to avoid propagating any nans (those get
        # propagated through the cond even if the spiking version isn't
        # being used at all)
        return tf.stop_gradient(spikes), tf.stop_gradient(voltage)



import time
import timeit

import nengo
import numpy as np
import pytest

from nengo_extras.simulators import RealTimeSimulator


def test_real_time_simulator(rng):
    with nengo.Network() as net:

        def rnd_delay(_):
            delay = rng.uniform(0, 5e-4)
            start = timeit.default_timer()
            while timeit.default_timer() - start < delay:
                pass
            return [0.0]

        nengo.Node(rnd_delay)

    with RealTimeSimulator(net) as sim:
        for _ in range(2):
            t_start = timeit.default_timer()
            sim.run(1)
            elapsed = timeit.default_timer() - t_start

            assert np.allclose(elapsed, 1, atol=0.02)

            # verify that resetting can be used to address external delays
            time.sleep(0.1)
            sim.reset()

    # warning when running slower than real-time
    with nengo.Network() as net:

        def long_delay(_):
            time.sleep(0.1)

        nengo.Node(long_delay)

    with RealTimeSimulator(net) as sim:
        with pytest.warns(UserWarning, match="RealTimeSimulator is running"):
            sim.step()



import nengo

from . import reqs
from .version import version as __version__


def set_rc_default(rc, section, option, value=None):
    if not rc.has_section(section):
        rc.add_section(section)
    if not rc.has_option(section, option):
        rc.set(section, option, value=value)


set_rc_default(nengo.rc, "nengo_extras", "data_dir", ".")


__copyright__ = "2015-2021, Applied Brain Research"
__license__ = "Free for non-commercial use; see LICENSE.rst"



import time
import timeit
import warnings

import nengo


class RealTimeSimulator(nengo.Simulator):
    """
    A simulator that will not run faster than real time.

    Nothing prevents this simulator from running slower than real time
    (although it will print a warning if that is the case).

    In addition, this Simulator measures time from the first call to
    ``sim.step``/``sim.run``, so adding some large gap in the middle of a
    run will throw off the timing.  Calling ``sim.reset()`` will reset the
    start time.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.start_time = None

    def step(self, *args, **kwargs):
        if self.start_time is None:
            self.start_time = timeit.default_timer()

        super().step(*args, **kwargs)

        elapsed = timeit.default_timer() - self.start_time

        if elapsed < self.time:
            time.sleep(self.time - elapsed)
        elif elapsed - self.time > 0.05:
            warnings.warn("RealTimeSimulator is running slower than real time")

    def reset(self, *args, **kwargs):
        super().reset(*args, **kwargs)

        self.start_time = None



"""
Tools for running deep networks in Nengo

Notes:
- Layers with weights make copies of said weights, both so that they are
  independent (if the model is later updated), and so they are contiguous.
  (Contiguity is required in Nengo when arrays are hashed during build.)
"""

from __future__ import absolute_import

import nengo
import numpy as np
from nengo.params import Default
from nengo.utils.network import with_self

from nengo_extras.convnet import Conv2d, Pool2d, softmax


class DeepNetwork(nengo.Network):
    @with_self
    def add_data_layer(self, d, name=None, **kwargs):
        kwargs.setdefault("label", name)
        layer = DataLayer(d, **kwargs)
        return self.add_layer(layer, inputs=(), name=name)

    @with_self
    def add_neuron_layer(self, n, inputs=None, name=None, **kwargs):
        kwargs.setdefault("label", name)
        layer = NeuronLayer(n, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_softmax_layer(self, size, inputs=None, name=None, **kwargs):
        kwargs.setdefault("label", name)
        layer = SoftmaxLayer(size, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_dropout_layer(self, size, keep, inputs=None, name=None, **kwargs):
        kwargs.setdefault("label", name)
        layer = DropoutLayer(size, keep, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_full_layer(self, weights, biases, inputs=None, name=None, **kwargs):
        kwargs.setdefault("label", name)
        layer = FullLayer(weights, biases, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_local_layer(
        self, input_shape, filters, biases, inputs=None, name=None, **kwargs
    ):
        kwargs.setdefault("label", name)
        layer = LocalLayer(input_shape, filters, biases, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_conv_layer(
        self, input_shape, filters, biases, inputs=None, name=None, **kwargs
    ):
        kwargs.setdefault("label", name)
        layer = ConvLayer(input_shape, filters, biases, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_pool_layer(self, input_shape, pool_size, inputs=None, name=None, **kwargs):
        kwargs.setdefault("label", name)
        layer = PoolLayer(input_shape, pool_size, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    def compute(self, inputs, output):
        raise NotImplementedError()


class SequentialNetwork(DeepNetwork):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.layers = []
        self.layers_by_name = {}

    @property
    def input(self):
        return None if len(self.layers) == 0 else self.layers[0].input

    @property
    def output(self):
        return None if len(self.layers) == 0 else self.layers[-1].output

    @with_self
    def add_layer(self, layer, inputs=None, name=None):
        assert isinstance(layer, Layer)
        assert layer not in self.layers

        if inputs is None:
            inputs = [self.layers[-1]] if self.output is not None else []

        assert (
            len(inputs) == 0
            if self.output is None
            else (len(inputs) == 1 and inputs[0] is self.layers[-1])
        )

        for i in inputs:
            nengo.Connection(i.output, layer.input, synapse=None, **layer.pre_args)
        self.layers.append(layer)

        if name is not None:
            assert name not in self.layers_by_name
            self.layers_by_name[name] = layer

        return layer

    def layers_to(self, end):
        if end is None:
            return self.layers

        assert end in self.layers
        return self.layers[: self.layers.index(end) + 1]

    def compute(self, inputs, output=None):
        y = inputs
        for layer in self.layers_to(output):
            y = layer.compute(y)

        return y


class Layer(nengo.Network):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pre_args = {}  # kwargs for incoming connections
        # self.post_args = {}  # kwargs for outgoing connections

    @property
    def input(self):
        raise NotImplementedError()

    @property
    def output(self):
        raise NotImplementedError()

    @property
    def size_in(self):
        return self.input.size_in

    @property
    def size_out(self):
        return self.output.size_out

    @property
    def shape_in(self):
        return (self.size_in,)

    @property
    def shape_out(self):
        return (self.size_out,)

    def compute(self, x):
        raise NotImplementedError()

    def _compute_input(self, x):
        x = np.asarray(x)
        if x.ndim == 0:
            raise ValueError("'x' must be at least 1D")
        if x.ndim == 1:
            assert x.shape[0] == self.size_in
            return x.reshape((1, -1))

        assert x.ndim == 2 and x.shape[1] == self.size_in
        return x


class NodeLayer(Layer):
    def __init__(self, output=None, size_in=None, **kwargs):
        super().__init__(**kwargs)

        with self:
            self.node = nengo.Node(output=output, size_in=size_in)

        if self.label is not None:
            self.node.label = "%s_node" % self.label

    @property
    def input(self):
        return self.node

    @property
    def output(self):
        return self.node

    def compute(self, x):
        raise NotImplementedError()


class NeuronLayer(Layer):
    def __init__(
        self,
        n,
        neuron_type=Default,
        synapse=Default,
        gain=1.0,
        bias=0.0,
        amplitude=1.0,
        **kwargs,
    ):
        super().__init__(**kwargs)

        with self:
            self.ensemble = nengo.Ensemble(n, 1, neuron_type=neuron_type)
            self.ensemble.gain = _vectorize(gain, n, "gain")
            self.ensemble.bias = _vectorize(bias, n, "bias")

            self._output = nengo.Node(size_in=n)
            self.connection = nengo.Connection(
                self.ensemble.neurons, self.output, transform=amplitude, synapse=synapse
            )

        if self.label is not None:
            self.ensemble.label = "%s_ensemble" % self.label
            self.output.label = "%s_output" % self.label

    @property
    def input(self):
        return self.ensemble.neurons

    @property
    def output(self):
        return self._output

    @property
    def neurons(self):
        return self.ensemble.neurons

    @property
    def amplitude(self):
        return self.connection.transform

    @property
    def bias(self):
        return self.ensemble.bias

    @property
    def gain(self):
        return self.ensemble.gain

    @property
    def neuron_type(self):
        return self.ensemble.neuron_type

    @property
    def synapse(self):
        return self.connection.synapse

    def compute(self, x):
        x = self._compute_input(x)
        return self.amplitude * self.neuron_type.rates(x, self.gain, self.bias)


class DataLayer(NodeLayer):
    def __init__(self, size, **kwargs):
        super().__init__(size_in=size, **kwargs)

    def compute(self, x):
        return x.reshape((x.shape[0], self.size_out))


class SoftmaxLayer(NodeLayer):
    def __init__(self, size, **kwargs):
        super().__init__(output=lambda t, x: softmax(x), size_in=size, **kwargs)

    def compute(self, x):
        x = self._compute_input(x)
        return softmax(x, axis=-1)


class DropoutLayer(NodeLayer):
    def __init__(self, size, keep, **kwargs):
        super().__init__(size_in=size, **kwargs)
        self.pre_args["transform"] = keep

    @property
    def keep(self):
        return self.pre_args["transform"]

    def compute(self, x):
        x = self._compute_input(x)
        return self.keep * x


class FullLayer(NodeLayer):
    def __init__(self, weights, biases, **kwargs):
        assert weights.ndim == 2
        assert biases.size == weights.shape[0]
        super().__init__(size_in=weights.shape[0], **kwargs)

        self.weights = np.array(weights)  # copy
        self.biases = np.array(biases)  # copy

        with self:
            self.bias = nengo.Node(output=biases)
            nengo.Connection(self.bias, self.node, synapse=None)

        self.pre_args["transform"] = weights

        if self.label is not None:
            self.bias.label = "%s_bias" % self.label

    def compute(self, x):
        x = self._compute_input(x)
        return np.dot(x, self.weights.T) + self.biases


class ProcessLayer(NodeLayer):
    def __init__(self, process, **kwargs):
        assert isinstance(process, nengo.Process)
        super().__init__(output=process, **kwargs)
        self._step = None

    @property
    def process(self):
        return self.node.output

    @property
    def shape_in(self):
        return self.process.shape_in

    @property
    def shape_out(self):
        return self.process.shape_out

    def compute(self, x):
        x = self._compute_input(x)
        n = x.shape[0]
        if self._step is None:
            self._step = self.process.make_step(
                self.size_in, self.size_out, dt=1.0, rng=None, state={}
            )
        return self._step(0, x).reshape((n, -1))


class LocalLayer(ProcessLayer):
    def __init__(self, input_shape, filters, biases, strides=1, padding=0, **kwargs):
        assert filters.ndim == 6
        filters = np.array(filters)  # copy
        biases = np.array(biases)  # copy
        p = Conv2d(input_shape, filters, biases, strides=strides, padding=padding)
        super().__init__(p, **kwargs)


class ConvLayer(ProcessLayer):
    def __init__(
        self,
        input_shape,
        filters,
        biases,
        strides=1,
        padding=0,
        border="ceil",
        **kwargs,
    ):
        assert filters.ndim == 4
        filters = np.array(filters)  # copy
        biases = np.array(biases)  # copy
        p = Conv2d(
            input_shape,
            filters,
            biases,
            strides=strides,
            padding=padding,
            border=border,
        )
        super().__init__(p, **kwargs)


class PoolLayer(ProcessLayer):
    def __init__(
        self, input_shape, pool_size, strides=None, kind="avg", mode="full", **kwargs
    ):
        p = Pool2d(input_shape, pool_size, strides=strides, kind=kind, mode=mode)
        super().__init__(p, **kwargs)


def _vectorize(value, n, name):
    value = np.asarray(value)
    if value.size == 1:
        return value * np.ones(n)
    if value.size == n:
        return value
    raise ValueError("%r must be length 1 or %d (got %d)" % (name, n, value.size))



from __future__ import absolute_import

import os

import nengo
import numpy as np

import nengo_extras.deepnetworks
from nengo_extras import reqs
from nengo_extras.neurons import SoftLIFRate

if reqs.HAS_TENSORFLOW:
    import tensorflow as tf

if reqs.HAS_KERAS:
    import keras
    from keras import backend as K
    from keras.layers import Convolution2D, Layer, LocallyConnected2D
else:

    class Layer:
        pass


class SoftLIF(Layer):
    def __init__(
        self,
        sigma=1.0,
        amplitude=1.0,
        tau_rc=0.02,
        tau_ref=0.002,
        noise_model="none",
        tau_s=0.005,
        **kwargs,
    ):
        if not reqs.HAS_KERAS:
            raise ImportError("`SoftLIF` requires `keras`")

        self.supports_masking = True
        self.sigma = sigma
        self.amplitude = amplitude
        self.tau_rc = tau_rc
        self.tau_ref = tau_ref

        if noise_model not in ("none", "alpharc"):
            raise ValueError("Unrecognized noise model")
        self.noise_model = noise_model

        self.tau_s = tau_s
        if abs(self.tau_rc - self.tau_s) < 1e-4:
            raise ValueError("tau_rc and tau_s must be different")

        super().__init__(**kwargs)

    def call(self, x, mask=None):
        """Compute the SoftLIF nonlinearity."""

        if not reqs.HAS_TENSORFLOW:
            raise ImportError('Keras backend is "tensorflow" but it cannot be imported')

        assert K.backend() == "tensorflow"

        # compute non-noisy output
        xs = x / self.sigma
        x_valid = xs > -20
        xs_safe = tf.where(x_valid, xs, K.zeros_like(xs))
        j = K.softplus(xs_safe) * self.sigma
        p = self.tau_ref + self.tau_rc * tf.where(
            x_valid, tf.math.log1p(1 / j), -xs - np.log(self.sigma)
        )
        r = self.amplitude / p

        if self.noise_model == "none":
            return r

        assert self.noise_model == "alpharc"
        # compute noisy output for forward pass
        d = self.tau_rc - self.tau_s
        u01 = K.random_uniform(K.shape(p))
        t = u01 * p
        q_rc = K.exp(-t / self.tau_rc)
        q_s = K.exp(-t / self.tau_s)
        r_rc1 = -tf.math.expm1(-p / self.tau_rc)  # 1 - exp(-p/tau_rc)
        r_s1 = -tf.math.expm1(-p / self.tau_s)  # 1 - exp(-p/tau_s)

        pt = tf.where(p < 100 * self.tau_s, (p - t) * (1 - r_s1), K.zeros_like(p))
        qt = tf.where(t < 100 * self.tau_s, q_s * (t + pt), K.zeros_like(t))
        rt = qt / (self.tau_s * d * r_s1 ** 2)
        rn = self.tau_rc * (q_rc / (d * d * r_rc1) - q_s / (d * d * r_s1)) - rt

        # r + stop_gradient(rn - r) = rn on forward pass, r on backwards
        return r + K.stop_gradient(self.amplitude * rn - r)

    def get_config(self):
        """Return a config dict to reproduce this SoftLIF."""
        config = {
            "sigma": self.sigma,
            "amplitude": self.amplitude,
            "tau_rc": self.tau_rc,
            "tau_ref": self.tau_ref,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


_custom_objects = {"SoftLIF": SoftLIF} if reqs.HAS_KERAS else {}


def load_model_pair(filepath, custom_objects=None):
    if custom_objects is None:
        custom_objects = {}

    json_path = filepath + ".json"
    h5_path = filepath + ".h5"

    combined_customs = dict(_custom_objects)
    combined_customs.update(custom_objects)

    with open(json_path, "r", encoding="utf-8") as f:
        model = keras.models.model_from_json(f.read(), custom_objects=combined_customs)

    model.load_weights(h5_path)
    return model


def save_model_pair(model, filepath, overwrite=False):
    json_path = filepath + ".json"
    h5_path = filepath + ".h5"

    if not overwrite and os.path.exists(json_path):
        raise ValueError("Path already exists: %r" % filepath)

    json_string = model.to_json()
    with open(json_path, "w", encoding="utf-8") as f:
        f.write(json_string)

    model.save_weights(h5_path, overwrite=overwrite)


def kmodel_compute_shapes(kmodel, input_shape):
    assert isinstance(kmodel, keras.models.Sequential)

    shapes = [input_shape]
    for layer in kmodel.layers:
        s = layer.compute_output_shape(shapes[-1])
        shapes.append(s)

    return shapes


class SequentialNetwork(nengo_extras.deepnetworks.SequentialNetwork):
    def __init__(self, model, synapse=None, lif_type="lif", **kwargs):
        super().__init__(**kwargs)

        assert isinstance(model, keras.models.Sequential)
        self.model = model
        self.synapse = synapse
        self.lif_type = lif_type

        # -- build model
        self.add_data_layer(np.prod(model.input_shape[1:]))
        for layer in model.layers:
            self._add_layer(layer)

    def _add_layer(self, layer):
        assert layer.input_mask is None
        assert layer.input_shape[0] is None

        layer_adder = {
            keras.layers.Activation: self._add_activation_layer,
            keras.layers.Dense: self._add_dense_layer,
            keras.layers.Dropout: self._add_dropout_layer,
            keras.layers.Flatten: self._add_flatten_layer,
            keras.layers.Convolution2D: self._add_conv2d_layer,
            keras.layers.AveragePooling2D: self._add_avgpool2d_layer,
            keras.layers.MaxPooling2D: self._add_maxpool2d_layer,
            keras.layers.noise.GaussianNoise: self._add_gaussian_noise_layer,
            SoftLIF: self._add_softlif_layer,
        }

        for cls in type(layer).__mro__:
            if cls in layer_adder:
                return layer_adder[cls](layer)

        raise NotImplementedError("Cannot build layer type %r" % type(layer).__name__)

    def _add_dense_layer(self, layer):
        weights, biases = layer.get_weights()
        return self.add_full_layer(weights.T, biases, name=layer.name)

    def _add_conv2d_layer(self, layer):
        if not reqs.HAS_KERAS:
            raise ImportError("`_add_conv2d_layer` requires `keras`")

        shape_in = layer.input_shape[1:]
        filters, biases = layer.get_weights()
        strides = layer.strides

        assert layer.data_format == "channels_first"
        nc, _, _ = shape_in
        filters = np.transpose(filters, (3, 2, 0, 1))
        assert K.backend() == "tensorflow"
        filters = filters.copy()  # to make contiguous

        _, nc2, si, sj = filters.shape
        assert nc == nc2, "Filter channels must match input channels"

        if layer.padding == "valid":
            padding = (0, 0)
        elif layer.padding == "same":
            padding = ((si - 1) / 2, (sj - 1) / 2)
        else:
            raise ValueError("Unrecognized padding %r" % layer.padding)

        conv = self.add_conv_layer(
            shape_in,
            filters,
            biases,
            strides=strides,
            padding=padding,
            border="floor",
            name=layer.name,
        )
        assert conv.size_out == np.prod(layer.output_shape[1:])
        return conv

    def _add_pool2d_layer(self, layer, kind=None):
        shape_in = layer.input_shape[1:]
        pool_size = layer.pool_size
        strides = layer.strides
        return self.add_pool_layer(
            shape_in,
            pool_size,
            strides=strides,
            kind=kind,
            mode="valid",
            name=layer.name,
        )

    def _add_avgpool2d_layer(self, layer):
        return self._add_pool2d_layer(layer, kind="avg")

    def _add_maxpool2d_layer(self, layer):
        return self._add_pool2d_layer(layer, kind="max")

    def _add_activation_layer(self, layer):
        if layer.activation is keras.activations.softmax:
            return self._add_softmax_layer(layer)

        # add normal activation layer
        activation_map = {
            keras.activations.relu: nengo.neurons.RectifiedLinear(),
            keras.activations.sigmoid: nengo.neurons.Sigmoid(),
        }
        neuron_type = activation_map.get(layer.activation, None)
        if neuron_type is None:
            raise ValueError("Unrecognized activation type %r" % layer.activation)

        n = np.prod(layer.input_shape[1:])
        return self.add_neuron_layer(
            n,
            neuron_type=neuron_type,
            synapse=self.synapse,
            gain=1,
            bias=0,
            name=layer.name,
        )

    def _add_softlif_layer(self, layer):

        taus = dict(tau_rc=layer.tau_rc, tau_ref=layer.tau_ref)
        lif_type = self.lif_type.lower()
        if lif_type == "lif":
            neuron_type = nengo.LIF(**taus)
        elif lif_type == "lifrate":
            neuron_type = nengo.LIFRate(**taus)
        elif lif_type == "softlifrate":
            neuron_type = SoftLIFRate(sigma=layer.sigma, **taus)
        else:
            raise KeyError("Unrecognized LIF type %r" % self.lif_type)

        n = np.prod(layer.input_shape[1:])
        return self.add_neuron_layer(
            n,
            neuron_type=neuron_type,
            synapse=self.synapse,
            gain=1,
            bias=1,
            amplitude=layer.amplitude,
            name=layer.name,
        )

    def _add_softmax_layer(self, layer):
        return None  # non-neural, we can do without it
        # return self.add_softmax_layer(
        #     np.prod(layer.input_shape[1:]), name=layer.name)

    def _add_dropout_layer(self, layer):
        return None  # keras scales by dropout rate, so we don't have to

    def _add_flatten_layer(self, layer):
        return None  # no computation, just reshaping, so ignore

    def _add_gaussian_noise_layer(self, layer):
        return None  # no noise during testing


def LSUVinit(kmodel, X, tol=0.1, t_max=50):  # noqa: C901
    """Layer-sequential unit-variance initialization.

    References
    ----------
    .. [1] Mishkin, D., & Matas, J. (2016). All you need is a good init.
       In ICLR 2016 (pp. 1-13).
    """
    if not reqs.HAS_KERAS:
        raise ImportError("`LSUVinit` requires `keras`")

    # f = K.function([kmodel.layers[0].input, K.learning_phase()],
    #                [klayer.output])
    # --- orthogonalize weights
    def orthogonalize(X):
        assert X.ndim == 2
        U, _, V = np.linalg.svd(X, full_matrices=False)
        return np.dot(U, V)

    for layer in kmodel.layers:
        weights = layer.get_weights()
        if len(weights) == 0:
            continue

        W, b = weights
        if isinstance(layer, Convolution2D):
            Wv = W.reshape((W.shape[0], -1))
        elif isinstance(layer, LocallyConnected2D):
            Wv = W.reshape((-1, W.shape[-1]))
        else:
            assert W.ndim == 2
            Wv = W

        Wv[:] = orthogonalize(Wv)
        layer.set_weights((W, b))

    # --- adjust variances
    s_input = kmodel.layers[0].input
    for layer in kmodel.layers:
        weights = layer.get_weights()
        if len(weights) == 0:
            continue

        W, b = weights
        f = K.function([s_input, K.learning_phase()], [layer.output])
        learning_phase = 0  # 0 == testing, 1 == training

        for _ in range(t_max):
            Y = f([X, learning_phase])[0]
            Ystd = Y.std()
            print(Ystd)
            if abs(Ystd - 1) < tol:
                break

            W /= Ystd
            layer.set_weights((W, b))
        else:
            print(
                "Layer %r did not converge after %d iterations (Ystd=%0.3e)"
                % (layer.name, t_max, Ystd)
            )



# pylint: disable=broad-except,unused-import

import logging

logger = logging.getLogger(__name__)

try:
    import tensorflow

    HAS_TENSORFLOW = True
except Exception as err:
    HAS_TENSORFLOW = False
    logger.debug("Error importing TensorFlow:\n%s", err)

try:
    import nengo_dl.neuron_builders

    HAS_NENGO_DL = True
    assert HAS_TENSORFLOW, "NengoDL installed without Tensorflow"
except Exception as err:
    HAS_NENGO_DL = False
    logger.debug("Error importing NengoDL:\n%s", err)


try:
    import nengo_loihi

    HAS_NENGO_LOIHI = True
except Exception as err:
    HAS_NENGO_LOIHI = False
    logger.debug("Error importing NengoLoihi:\n%s", err)

try:
    import gi

    gi.require_version("Gst", "1.0")
    import gi.repository

    HAS_GI = True
except Exception as err:
    HAS_GI = False
    logger.debug("Error importing PyGObject:\n%s", err)


try:
    import scipy.cluster.hierarchy
    import scipy.interpolate
    import scipy.ndimage.interpolation
    import scipy.stats

    HAS_SCIPY = True
except Exception as err:
    HAS_SCIPY = False
    logger.debug("Error importing SciPy:\n%s", err)


try:
    import keras.backend
    import keras.layers

    HAS_KERAS = True
except Exception as err:
    HAS_KERAS = False
    logger.debug("Error importing Keras:\n%s", err)


try:
    import numba.extending

    HAS_NUMBA = True
except Exception as err:
    HAS_NUMBA = False
    logger.debug("Error importing Numba:\n%s", err)



import re
import xml.etree.ElementTree as et

import nengo
import pytest

from nengo_extras.gexf import (
    CollapsingGexfConverter,
    DispatchTable,
    GexfConverter,
    HierarchicalLabeler,
)


def test_can_dispatch_table_defaults():
    class A:
        def __init__(self):
            self.called = False

    class Test:
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = True

    a = A()
    Test().dispatch(a)
    assert a.called


def test_dispatch_obj_inheritance():
    class A:
        def __init__(self):
            self.called = False

    class B(A):
        pass

    class C(A):
        pass

    class Test:
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = "a"

        @dispatch.register(B)
        def process_b(self, obj):
            obj.called = "b"

    a = A()
    Test().dispatch(a)
    assert a.called == "a"

    b = B()
    Test().dispatch(b)
    assert b.called == "b"

    c = C()
    Test().dispatch(c)
    assert c.called == "a"


def test_dispatch_cls_inheritance():
    class A:
        def __init__(self):
            self.called = False

    class B:
        def __init__(self):
            self.called = False

    class C:
        def __init__(self):
            self.called = False

    class Test1:
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = "test1.process_a"

        @dispatch.register(B)
        def process_b(self, obj):
            obj.called = "test1.process_b"

    class Test2(Test1):
        dispatch = DispatchTable(Test1.dispatch)

        @dispatch.register(B)
        def process_b(self, obj):
            obj.called = "test2.process_b"

        @dispatch.register(C)
        def process_c(self, obj):
            obj.called = "test2.process_c"

    a = A()
    b = B()
    Test1().dispatch(a)
    assert a.called == "test1.process_a"
    Test1().dispatch(b)
    assert b.called == "test1.process_b"

    a = A()
    b = B()
    c = C()
    Test2().dispatch(a)
    assert a.called == "test1.process_a"
    Test2().dispatch(b)
    assert b.called == "test2.process_b"
    Test2().dispatch(c)
    assert c.called == "test2.process_c"


def test_dispatch_instance_specific():
    class A:
        def __init__(self):
            self.called = False

    class B:
        def __init__(self):
            self.called = False

    class Test:
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = "test.process_a"

    def proc_inst(obj):
        obj.called = "proc_inst"

    test = Test()
    test2 = Test()
    test.dispatch.register(A, proc_inst)
    test.dispatch.register(B, proc_inst)

    a = A()
    test.dispatch(a)
    assert a.called == "proc_inst"
    test2.dispatch(a)
    assert a.called == "test.process_a"

    b = B()
    test.dispatch(b)
    assert b.called == "proc_inst"


def test_dispatch_errors():
    class Test:
        dispatch = DispatchTable()

    with pytest.raises(NotImplementedError):
        Test().dispatch(object())


def test_hierarchical_labeler():
    with nengo.Network() as model:
        model.ens = nengo.Ensemble(10, 1)
        with nengo.Network() as model.subnet:
            model.subnet.node = nengo.Node(1.0)
            model.subnet.attr = nengo.Ensemble(10, 1)

    labels = HierarchicalLabeler().get_labels(model)
    assert dict(labels) == {
        model.ens: "ens",
        model.subnet: "subnet",
        model.subnet.node: "subnet.node",
        model.subnet.attr: "subnet.attr",
    }


@pytest.mark.xfail(reason="TODO: debug this failure")
def test_gexf_converter():
    with nengo.Network() as model:
        model.ens = nengo.Ensemble(10, 1)
        with nengo.Network() as model.subnet:
            model.subnet.node = nengo.Node(1.0)
            model.subnet.attr = nengo.Ensemble(10, 1)
        nengo.Connection(model.subnet.node, model.ens)
        nengo.Probe(model.ens)

    expected = (
        '<gexf version="1.3" '
        'xmlns="http://www.gexf.net/1.3draft" '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xsi:schemaLocation="http://www.gexf.net/1.3draft '
        'http://www.gexf.net/1.3draft/gexf.xsd">'
        '<meta lastmodifieddate="\\d+-\\d{2}-\\d{2}">'
        "<creator>nengo_extras.gexf.GexfConverter</creator>"
        "</meta>"
        '<graph defaultedgetype="directed">'
        '<attributes class="node">'
        '<attribute id="0" title="type" type="string" />'
        '<attribute id="1" title="net" type="long" />'
        '<attribute id="2" title="net_label" type="string" />'
        '<attribute id="3" title="size_in" type="integer" />'
        '<attribute id="4" title="size_out" type="integer" />'
        '<attribute id="5" title="radius" type="float" />'
        '<attribute id="6" title="n_neurons" type="integer">'
        "<default>0</default>"
        "</attribute>"
        '<attribute id="7" title="neuron_type" type="string" />'
        "</attributes>"
        '<attributes class="edge">'
        '<attribute id="0" title="pre_type" type="string" />'
        '<attribute id="1" title="post_type" type="string" />'
        '<attribute id="2" title="synapse" type="string" />'
        '<attribute id="3" title="tau" type="float" />'
        '<attribute id="4" title="function" type="string" />'
        '<attribute id="5" title="transform" type="string" />'
        '<attribute id="6" title="scalar_transform" type="float">'
        "<default>1.0</default>"
        "</attribute>"
        '<attribute id="7" title="learning_rule_type" type="string" />'
        "</attributes>"
        "<nodes>"
        '<node id="\\d+" label="ens">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        "</attvalues>"
        "</node>"
        '<node id="\\d+" label="subnet.attr">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        "</attvalues>"
        "</node>"
        '<node id="\\d+" label="subnet.node">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="0" />'
        '<attvalue for="4" value="1" />'
        "</attvalues>"
        "</node>"
        "</nodes>"
        "<edges>"
        '<edge id="\\d+" source="\\d+" target="\\d+">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="2" value="Lowpass\\(0.005\\)" />'
        '<attvalue for="3" value="0.005" />'
        '<attvalue for="5" value="1.0" />'
        "</attvalues>"
        "</edge>"
        "</edges>"
        "</graph>"
        "</gexf>$"
    )

    actual = et.tostring(GexfConverter().convert(model).getroot()).decode("utf-8")
    assert re.match(expected, actual), actual


@pytest.mark.xfail(reason="TODO: debug this failure")
def test_gexf_converter_hierarchical():
    with nengo.Network() as model:
        model.ens = nengo.Ensemble(10, 1)
        with nengo.Network() as model.subnet:
            model.subnet.node = nengo.Node(1.0)
            model.subnet.attr = nengo.Ensemble(10, 1)
        nengo.Connection(model.subnet.node, model.ens)
        nengo.Probe(model.ens)

    expected = (
        '<gexf version="1.3" '
        'xmlns="http://www.gexf.net/1.3draft" '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xsi:schemaLocation="http://www.gexf.net/1.3draft '
        'http://www.gexf.net/1.3draft/gexf.xsd">'
        '<meta lastmodifieddate="\\d+-\\d{2}-\\d{2}">'
        "<creator>nengo_extras.gexf.GexfConverter</creator>"
        "</meta>"
        '<graph defaultedgetype="directed">'
        '<attributes class="node">'
        '<attribute id="0" title="type" type="string" />'
        '<attribute id="1" title="net" type="long" />'
        '<attribute id="2" title="net_label" type="string" />'
        '<attribute id="3" title="size_in" type="integer" />'
        '<attribute id="4" title="size_out" type="integer" />'
        '<attribute id="5" title="radius" type="float" />'
        '<attribute id="6" title="n_neurons" type="integer">'
        "<default>0</default>"
        "</attribute>"
        '<attribute id="7" title="neuron_type" type="string" />'
        "</attributes>"
        '<attributes class="edge">'
        '<attribute id="0" title="pre_type" type="string" />'
        '<attribute id="1" title="post_type" type="string" />'
        '<attribute id="2" title="synapse" type="string" />'
        '<attribute id="3" title="tau" type="float" />'
        '<attribute id="4" title="function" type="string" />'
        '<attribute id="5" title="transform" type="string" />'
        '<attribute id="6" title="scalar_transform" type="float">'
        "<default>1.0</default>"
        "</attribute>"
        '<attribute id="7" title="learning_rule_type" type="string" />'
        "</attributes>"
        "<nodes>"
        '<node id="\\d+" label="ens">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        "</attvalues>"
        "</node>"
        '<node id="\\d+" label="subnet">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.network.Network" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="6" value="10" />'
        "</attvalues>"
        "<nodes>"
        '<node id="\\d+" label="subnet.attr">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        "</attvalues>"
        "</node>"
        '<node id="\\d+" label="subnet.node">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="0" />'
        '<attvalue for="4" value="1" />'
        "</attvalues>"
        "</node>"
        "</nodes>"
        "</node>"
        "</nodes>"
        "<edges>"
        '<edge id="\\d+" source="\\d+" target="\\d+">'
        "<attvalues>"
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="2" value="Lowpass\\(0.005\\)" />'
        '<attvalue for="3" value="0.005" />'
        '<attvalue for="5" value="1.0" />'
        "</attvalues>"
        "</edge>"
        "</edges>"
        "</graph>"
        "</gexf>$"
    )

    actual = et.tostring(
        GexfConverter(hierarchical=True).convert(model).getroot()
    ).decode("utf-8")
    assert re.match(expected, actual), actual


@pytest.mark.xfail(reason="TODO: debug this failure")
def test_collapsing_gexf_converter():
    with nengo.Network() as model:
        model.ea = nengo.networks.EnsembleArray(10, 10)

    expected = (
        '<gexf version="1.3" '
        'xmlns="http://www.gexf.net/1.3draft" '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xsi:schemaLocation="http://www.gexf.net/1.3draft '
        'http://www.gexf.net/1.3draft/gexf.xsd">'
        '<meta lastmodifieddate="\\d+-\\d{2}-\\d{2}">'
        "<creator>nengo_extras.gexf.CollapsingGexfConverter</creator>"
        "</meta>"
        '<graph defaultedgetype="directed">'
        '<attributes class="node">'
        '<attribute id="0" title="type" type="string" />'
        '<attribute id="1" title="net" type="long" />'
        '<attribute id="2" title="net_label" type="string" />'
        '<attribute id="3" title="size_in" type="integer" />'
        '<attribute id="4" title="size_out" type="integer" />'
        '<attribute id="5" title="radius" type="float" />'
        '<attribute id="6" title="n_neurons" type="integer">'
        "<default>0</default>"
        "</attribute>"
        '<attribute id="7" title="neuron_type" type="string" />'
        "</attributes>"
        '<attributes class="edge">'
        '<attribute id="0" title="pre_type" type="string" />'
        '<attribute id="1" title="post_type" type="string" />'
        '<attribute id="2" title="synapse" type="string" />'
        '<attribute id="3" title="tau" type="float" />'
        '<attribute id="4" title="function" type="string" />'
        '<attribute id="5" title="transform" type="string" />'
        '<attribute id="6" title="scalar_transform" type="float">'
        "<default>1.0</default>"
        "</attribute>"
        '<attribute id="7" title="learning_rule_type" type="string" />'
        "</attributes>"
        "<nodes>"
        '<node id="\\d+" label="ea">'
        "<attvalues>"
        '<attvalue for="0" value="'
        'nengo.networks.ensemblearray.EnsembleArray" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="6" value="100" />'
        "</attvalues>"
        "</node>"
        "</nodes>"
        "<edges>"
        ".*"
        "</edges>"
        "</graph>"
        "</gexf>$"
    )

    actual = et.tostring(CollapsingGexfConverter().convert(model).getroot()).decode(
        "utf-8"
    )
    assert re.match(expected, actual), actual



import nengo
import numpy as np
import pytest

from nengo_extras import reqs
from nengo_extras.loihi_training import AlphaRCNoise, LowpassRCNoise, NoiseBuilder
from nengo_extras.neurons import SoftLIFRate

if reqs.HAS_NENGO_DL:
    import nengo_dl

if reqs.HAS_NENGO_LOIHI:
    from nengo_loihi.neurons import (
        LoihiLIF,
        LoihiSpikingRectifiedLinear,
        discretize_tau_rc,
        discretize_tau_ref,
        install_dl_builders,
        loihi_rates,
        nengo_rates,
    )
else:

    LoihiLIF = LoihiSpikingRectifiedLinear = lambda *args, **kwargs: None


def test_noisebuilder_register_twice_warning():
    class MockNoiseBuilder:
        pass

    # First time does not warn
    with pytest.warns(None) as record:
        NoiseBuilder.register(int)(MockNoiseBuilder)
    assert len(record) == 0

    # Second time warns
    with pytest.warns(UserWarning, match="already has a builder"):
        NoiseBuilder.register(int)(MockNoiseBuilder)


@pytest.mark.skipif(not reqs.HAS_NENGO_DL, reason="requires nengo-dl")
@pytest.mark.skipif(not reqs.HAS_NENGO_LOIHI, reason="requires nengo-loihi")
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.filterwarnings("ignore:invalid value")
@pytest.mark.filterwarnings("ignore:Non-finite values detected")
@pytest.mark.parametrize("neuron_type", [LoihiLIF(), LoihiSpikingRectifiedLinear()])
@pytest.mark.parametrize("inference_only", (True, False))
def test_nengo_dl_neurons(neuron_type, inference_only, Simulator, plt, allclose):
    install_dl_builders()

    dt = 0.0007

    n = 256
    encoders = np.ones((n, 1))
    gain = np.zeros(n)
    if isinstance(neuron_type, nengo.SpikingRectifiedLinear):
        bias = np.linspace(0, 1001, n)
    else:
        bias = np.linspace(0, 30, n)

    with nengo.Network() as model:
        nengo_dl.configure_settings(inference_only=inference_only)

        a = nengo.Ensemble(
            n, 1, neuron_type=neuron_type, encoders=encoders, gain=gain, bias=bias
        )
        ap = nengo.Probe(a.neurons)

    t_final = 1.0
    with nengo_dl.Simulator(model, dt=dt) as dl_sim:
        dl_sim.run(t_final)

    with Simulator(model, dt=dt) as loihi_sim:
        loihi_sim.run(t_final)

    rates_dlsim = (dl_sim.data[ap] > 0).sum(axis=0) / t_final
    rates_loihisim = (loihi_sim.data[ap] > 0).sum(axis=0) / t_final

    zeros = np.zeros((1, gain.size))
    rates_ref = neuron_type.rates(zeros, gain, bias, dt=dt).squeeze(axis=0)
    plt.plot(bias, rates_loihisim, "r", label="loihi sim")
    plt.plot(bias, rates_dlsim, "b-.", label="dl sim")
    plt.plot(bias, rates_ref, "k--", label="rates_ref")
    plt.legend(loc="best")

    atol = 1.0 / t_final  # the fundamental unit for our rates
    assert rates_ref.shape == rates_dlsim.shape == rates_loihisim.shape
    assert allclose(rates_dlsim, rates_ref, atol=atol, rtol=0, xtol=1)
    assert allclose(rates_loihisim, rates_ref, atol=atol, rtol=0, xtol=1)


def rate_nengo_dl_net(
    neuron_type, discretize=True, dt=0.001, nx=256, gain=1.0, bias=0.0
):
    """Create a network for determining rate curves with NengoDL.

    Arguments
    ---------
    neuron_type : NeuronType
        The neuron type used in the network's ensemble.
    discretize : bool, optional (Default: True)
        Whether the tau_ref and tau_rc values should be discretized
        before generating rate curves.
    dt : float, optional (Default: 0.001)
        Simulator timestep.
    nx : int, optional (Default: 256)
        Number of x points in the rate curve.
    gain : float, optional (Default: 1.)
        Gain of all neurons.
    bias : float, optional (Default: 0.)
        Bias of all neurons.
    """
    net = nengo.Network()
    net.dt = dt
    net.bias = bias
    net.gain = gain
    lif_kw = dict(amplitude=neuron_type.amplitude)
    if isinstance(neuron_type, LoihiLIF):
        net.x = np.linspace(-1, 30, nx)

        net.sigma = 0.02
        lif_kw["tau_rc"] = neuron_type.tau_rc
        lif_kw["tau_ref"] = neuron_type.tau_ref

        if discretize:
            lif_kw["tau_ref"] = discretize_tau_ref(lif_kw["tau_ref"], dt)
            lif_kw["tau_rc"] = discretize_tau_rc(lif_kw["tau_rc"], dt)
        lif_kw["tau_ref"] += 0.5 * dt

    elif isinstance(neuron_type, LoihiSpikingRectifiedLinear):
        net.x = np.linspace(-1, 999, nx)

        net.tau_ref1 = 0.5 * dt
        net.j = (
            neuron_type.current(net.x[:, np.newaxis], gain, bias).squeeze(axis=1) - 1
        )

    with net:
        if isinstance(neuron_type, LoihiLIF) and discretize:
            nengo_dl.configure_settings(lif_smoothing=net.sigma)

        net.stim = nengo.Node(np.zeros(nx))
        net.ens = nengo.Ensemble(
            nx,
            1,
            neuron_type=neuron_type,
            gain=nengo.dists.Choice([gain]),
            bias=nengo.dists.Choice([bias]),
        )
        nengo.Connection(net.stim, net.ens.neurons, synapse=None)
        net.probe = nengo.Probe(net.ens.neurons)

    rates = dict(ref=loihi_rates(neuron_type, net.x, gain, bias, dt=dt).squeeze(axis=1))
    # rates['med'] is an approximation of the smoothed Loihi tuning curve
    if isinstance(neuron_type, LoihiLIF):
        rates["med"] = nengo_rates(nengo.LIF(**lif_kw), net.x, gain, bias).squeeze(
            axis=1
        )
    elif isinstance(neuron_type, LoihiSpikingRectifiedLinear):
        rates["med"] = np.zeros_like(net.j)
        rates["med"][net.j > 0] = neuron_type.amplitude / (
            net.tau_ref1 + 1.0 / net.j[net.j > 0]
        )

    return net, rates, lif_kw


@pytest.mark.skipif(not reqs.HAS_NENGO_DL, reason="requires nengo-dl")
@pytest.mark.skipif(not reqs.HAS_NENGO_LOIHI, reason="requires nengo-loihi")
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.parametrize(
    "neuron_type",
    [
        LoihiLIF(amplitude=1.0, tau_rc=0.02, tau_ref=0.002),
        LoihiLIF(amplitude=0.063, tau_rc=0.05, tau_ref=0.001),
        LoihiSpikingRectifiedLinear(),
        LoihiSpikingRectifiedLinear(amplitude=0.42),
    ],
)
def test_nengo_dl_neuron_grads(neuron_type, plt, allclose):
    pytest.importorskip("tensorflow")

    install_dl_builders()

    net, rates, lif_kw = rate_nengo_dl_net(neuron_type)

    with net:
        nengo_dl.configure_settings(
            # run with `training=True`
            learning_phase=True,
            dtype="float64",
        )

    with nengo_dl.Simulator(net, dt=net.dt) as sim:
        sim.run_steps(1, data={net.stim: net.x[None, None, :]})
        y = sim.data[net.probe][0]

    # --- compute spiking rates
    n_spike_steps = 1000
    x_spikes = net.x + np.zeros((1, n_spike_steps, 1), dtype=net.x.dtype)
    with nengo_dl.Simulator(net, dt=net.dt) as sim:
        sim.run_steps(n_spike_steps, data={net.stim: x_spikes})
        y_spikes = sim.data[net.probe]
        y_spikerate = y_spikes.mean(axis=0)

    # --- compute derivatives
    if isinstance(neuron_type, LoihiLIF):
        dy_ref = SoftLIFRate(sigma=net.sigma, **lif_kw).derivative(
            net.x, net.gain, net.bias
        )
    else:
        # use the derivative of rates['med'] (the smoothed Loihi tuning curve)
        dy_ref = np.zeros_like(net.j)
        dy_ref[net.j > 0] = (
            neuron_type.amplitude / (net.j[net.j > 0] * net.tau_ref1 + 1) ** 2
        )

    with nengo_dl.Simulator(net, dt=net.dt) as sim:
        assert sim.unroll == 1
        # note: not actually checking gradients, just using this to get the gradients
        analytic = sim.check_gradients(inputs=[net.x[None, None, :]], atol=1e10)[
            net.probe
        ]["analytic"][0]

        dy = np.diagonal(analytic.copy())

    dx = net.x[1] - net.x[0]
    dy_est = np.diff(nengo.synapses.Alpha(10).filtfilt(rates["ref"], dt=1)) / dx
    x1 = 0.5 * (net.x[:-1] + net.x[1:])

    # --- plots
    plt.subplot(211)
    plt.plot(net.x, rates["med"], "--", label="LIF(tau_ref += 0.5*dt)")
    plt.plot(net.x, y, label="nengo_dl")
    plt.plot(net.x, y_spikerate, label="nengo_dl spikes")
    plt.plot(net.x, rates["ref"], "k--", label="LoihiLIF")
    plt.legend(loc=4)

    plt.subplot(212)
    plt.plot(x1, dy_est, "--", label="diff(smoothed_y)")
    plt.plot(net.x, dy, label="nengo_dl")
    plt.plot(net.x, dy_ref, "k--", label="diff(SoftLIF)")
    plt.legend(loc=1)

    np.fill_diagonal(analytic, 0)
    assert np.all(analytic == 0)

    assert y.shape == rates["ref"].shape
    assert allclose(y, rates["ref"], atol=1e-3, rtol=1e-5)
    assert allclose(dy, dy_ref, atol=1e-3, rtol=1e-5)
    assert allclose(y_spikerate, rates["ref"], atol=1, rtol=1e-2, record_rmse=False)


@pytest.mark.skipif(not reqs.HAS_NENGO_DL, reason="requires nengo-dl")
@pytest.mark.skipif(not reqs.HAS_NENGO_LOIHI, reason="requires nengo-loihi")
@pytest.mark.filterwarnings("ignore:divide by zero")
@pytest.mark.parametrize(
    "neuron_type",
    [
        LoihiLIF(amplitude=0.3, nengo_dl_noise=LowpassRCNoise(0.001)),
        LoihiLIF(amplitude=0.3, nengo_dl_noise=AlphaRCNoise(0.001)),
    ],
)
def test_nengo_dl_noise(neuron_type, seed, plt, allclose):
    pytest.importorskip("tensorflow")

    install_dl_builders()

    net, rates, _ = rate_nengo_dl_net(neuron_type)
    n_noise = 1000  # number of noise samples per x point

    with net:
        nengo_dl.configure_settings(learning_phase=True)  # run with `training=True`

    with nengo_dl.Simulator(net, dt=net.dt, minibatch_size=n_noise, seed=seed) as sim:
        input_data = {net.stim: np.tile(net.x[None, None, :], (n_noise, 1, 1))}
        sim.step(data=input_data)
        y = sim.data[net.probe][:, 0, :]

    ymean = y.mean(axis=0)
    y25 = np.percentile(y, 25, axis=0)
    y75 = np.percentile(y, 75, axis=0)
    dy25 = y25 - rates["ref"]
    dy75 = y75 - rates["ref"]

    # exponential models roughly fitted to 25/75th percentiles
    x1mask = net.x > 1.5
    x1 = net.x[x1mask]
    if isinstance(neuron_type.nengo_dl_noise, AlphaRCNoise):
        exp_model = 0.7 + 2.8 * np.exp(-0.22 * (x1 - 1))
        atol = 0.12 * exp_model.max()
    elif isinstance(neuron_type.nengo_dl_noise, LowpassRCNoise):
        exp_model = 1.5 + 2.2 * np.exp(-0.22 * (x1 - 1))
        atol = 0.2 * exp_model.max()

    rtol = 0.2
    mu_atol = 0.6  # depends on n_noise and variance of noise

    # --- plots
    plt.subplot(211)
    plt.plot(net.x, rates["med"], "--", label="LIF(tau_ref += 0.5*dt)")
    plt.plot(net.x, ymean, label="nengo_dl")
    plt.plot(net.x, y25, ":", label="25th")
    plt.plot(net.x, y75, ":", label="75th")
    plt.plot(net.x, rates["ref"], "k--", label="LoihiLIF")
    plt.legend()

    plt.subplot(212)
    plt.plot(net.x, ymean - rates["ref"], "b", label="mean")
    plt.plot(net.x, mu_atol * np.ones_like(net.x), "b:")
    plt.plot(net.x, -mu_atol * np.ones_like(net.x), "b:")
    plt.plot(net.x, y25 - rates["ref"], ":", label="25th")
    plt.plot(net.x, y75 - rates["ref"], ":", label="75th")
    plt.plot(x1, exp_model, "k--")
    plt.plot(x1, exp_model * (1 + rtol) + atol, "k:")
    plt.plot(x1, exp_model * (1 - rtol) - atol, "k:")
    plt.plot(x1, -exp_model, "k--")
    plt.legend()

    assert ymean.shape == rates["ref"].shape
    assert allclose(ymean, rates["ref"], atol=mu_atol, record_rmse=False)
    assert allclose(dy25[x1mask], -exp_model, atol=atol, rtol=rtol, record_rmse=False)
    assert allclose(dy75[x1mask], exp_model, atol=atol, rtol=rtol, record_rmse=False)



# pylint: disable=unused-import,ungrouped-imports

import sys

import nengo

if nengo.version.version_info < (3, 0, 0):
    from nengo.utils.compat import is_integer, is_iterable
else:
    from nengo.utils.numpy import is_integer, is_iterable

PY2 = sys.version_info[0] == 2

if PY2:
    from urllib import urlretrieve

    import cPickle as pickle
    import Tkinter as tkinter
    from cStringIO import StringIO
else:
    import pickle
    import tkinter
    from io import StringIO
    from urllib.request import urlretrieve


def cmp(a, b):  # same as python2's builtin cmp, not available in python3
    return (a > b) - (a < b)


def pickle_load(file, *args, **kwargs):
    if not PY2:
        kwargs.setdefault("encoding", "latin1")
    return pickle.load(file, *args, **kwargs)


def pickle_load_bytes(file, *args, **kwargs):
    if not PY2:
        kwargs.setdefault("encoding", "bytes")
    return pickle.load(file, *args, **kwargs)



import numpy as np
import pytest

from nengo_extras.data import (
    load_cifar10,
    load_cifar100,
    load_ilsvrc2012,
    load_mnist,
    load_svhn,
    one_hot_from_labels,
    spasafe_name,
    spasafe_names,
)
from nengo_extras.matplotlib import tile


@pytest.mark.slow
def test_load_cifar10(plt):
    trainX = load_cifar10(n_train=1, n_test=0)[0][0]
    trainX = trainX.reshape((-1, 3, 32, 32))
    trainX = np.transpose(trainX, (0, 2, 3, 1))
    tile(trainX, ax=plt.gca())


@pytest.mark.slow
def test_load_cifar100(plt):
    trainX = load_cifar100()[0][0]
    trainX = trainX.reshape((-1, 3, 32, 32))
    trainX = np.transpose(trainX, (0, 2, 3, 1))
    tile(trainX, ax=plt.gca())


@pytest.mark.slow
def test_load_ilsvrc2012(plt):
    testX, _, _, _ = load_ilsvrc2012(n_files=1)
    testX = testX.reshape((-1, 3, 256, 256))
    testX = np.transpose(testX, (0, 2, 3, 1))
    tile(testX, ax=plt.gca())


@pytest.mark.slow
def test_load_mnist(plt):
    trainX = load_mnist()[0][0]
    tile(trainX.reshape((-1, 28, 28)), ax=plt.gca())


@pytest.mark.slow
def test_load_svhn(plt):
    (trainX, _), (_, _) = load_svhn()
    trainX = trainX.reshape((-1, 3, 32, 32))
    trainX = np.transpose(trainX, (0, 2, 3, 1))
    tile(trainX, ax=plt.gca())


def test_one_hot_from_labels_int(rng):
    nc = 19
    labels = rng.randint(nc, size=1000)

    yref = np.zeros((len(labels), nc))
    yref[np.arange(len(labels)), labels] = 1

    y0 = one_hot_from_labels(labels)
    y1 = one_hot_from_labels(labels, classes=nc + 5)
    assert np.array_equal(y0, yref)
    assert np.array_equal(y0, y1[:, :nc])
    assert (y1[:, nc:] == 0).all()


def test_one_hot_from_labels_skip(rng):
    labels = 2 * rng.randint(4, size=1000)

    yref = np.zeros((len(labels), labels.max() + 1))
    yref[np.arange(len(labels)), labels] = 1
    y = one_hot_from_labels(labels)
    assert np.array_equal(y, yref)


def test_one_hot_from_labels_float(rng):
    classes = rng.uniform(0, 9, size=11)
    inds = rng.randint(len(classes), size=1000)
    labels = classes[inds]

    yref = np.zeros((len(labels), len(classes)))
    yref[np.arange(len(labels)), inds] = 1

    y = one_hot_from_labels(labels, classes=classes)
    assert np.array_equal(y, yref)


def test_spasafe_name():
    assert spasafe_name("UPPER") == "UPPER"
    assert spasafe_name("Camel") == "Camel"
    assert spasafe_name("lower") == "Lower"
    assert spasafe_name("Under_score") == "Under_score"
    assert (
        spasafe_name("Weird.,:[]^!<>=&'\"symbols", pre_comma_only=False)
        == "Weird_symbols"
    )
    assert spasafe_name("Weird.,:[]^!<>=&'\"symbols", pre_comma_only=True) == "Weird"
    assert spasafe_name("  \tWhite- \t\r\nspace  \r\n") == "White_space"
    assert spasafe_name("multiple correction's") == "Multiple_corrections"
    assert spasafe_name("123four") == "Four"
    assert spasafe_name("A") == "A"
    with pytest.raises(ValueError):
        spasafe_name("")


def test_spasafe_names():
    assert spasafe_names(["A,B", "A", "c's"]) == ["A0", "A1", "Cs"]



import os

import _pytest.capture
import pytest
from nengo.utils.stdlib import execfile

try:
    from nengo.utils.ipython import export_py, load_notebook
except ImportError as err:

    def export_py(*args, import_err=err, **kwargs):
        raise import_err

    def load_notebook(*args, import_err=err, **kwargs):
        raise import_err


install_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), os.pardir, os.pardir)
)
examples_dir = os.path.join(install_dir, "examples")

# Monkeypatch _pytest.capture.DontReadFromInput
#  If we don't do this, importing IPython will choke as it reads the current
#  sys.stdin to figure out the encoding it will use; pytest installs
#  DontReadFromInput as sys.stdin to capture output.
#  Running with -s option doesn't have this issue, but this monkeypatch
#  doesn't have any side effects, so it's fine.
_pytest.capture.DontReadFromInput.encoding = "utf-8"
_pytest.capture.DontReadFromInput.write = lambda: None
_pytest.capture.DontReadFromInput.flush = lambda: None


all_examples = []

for subdir, _, files in os.walk(examples_dir):
    if os.path.sep + "." in subdir:
        continue
    files = [f for f in files if f.endswith(".ipynb")]
    examples = [os.path.join(subdir, os.path.splitext(f)[0]) for f in files]
    all_examples.extend(examples)

# os.walk goes in arbitrary order, so sort after the fact to keep pytest happy
all_examples.sort()


def assert_noexceptions(nb_file, tmpdir):
    plt = pytest.importorskip("matplotlib.pyplot")
    pytest.importorskip("IPython", minversion="1.0")
    pytest.importorskip("jinja2")

    nb_path = os.path.join(examples_dir, "%s.ipynb" % nb_file)
    nb = load_notebook(nb_path)
    pyfile = "%s.py" % (tmpdir.join(os.path.splitext(os.path.basename(nb_path))[0]))
    export_py(nb, pyfile)
    execfile(pyfile, {})
    plt.close("all")


def iter_cells(nb_file, cell_type="code"):
    nb = load_notebook(os.path.join(examples_dir, "%s.ipynb" % nb_file))

    if nb.nbformat <= 3:
        cells = []
        for ws in nb.worksheets:
            cells.extend(ws.cells)
    else:
        cells = nb.cells

    for cell in cells:
        if cell.cell_type == cell_type:
            yield cell


@pytest.mark.example
@pytest.mark.parametrize("nb_file", all_examples)
def test_no_signature(nb_file):
    pytest.importorskip("IPython", minversion="1.0")

    nb = load_notebook(os.path.join(examples_dir, "%s.ipynb" % nb_file))
    assert "signature" not in nb.metadata, "Notebook has signature"


@pytest.mark.example
@pytest.mark.parametrize("nb_file", all_examples)
def test_no_outputs(nb_file):
    """Ensure that no cells have output."""
    pytest.importorskip("IPython", minversion="1.0")

    for cell in iter_cells(nb_file):
        assert cell.outputs == [], "Cell outputs not cleared"



#!/usr/bin/env python

# Automatically generated by nengo-bones, do not edit this file directly

import io
import pathlib
import runpy

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py"
    )


def read(*filenames, **kwargs):
    encoding = kwargs.get("encoding", "utf-8")
    sep = kwargs.get("sep", "\n")
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = pathlib.Path(__file__).parent
version = runpy.run_path(str(root / "nengo_extras" / "version.py"))["version"]

install_req = [
    "numpy>=1.8",
    "nengo>=3.0",
    "matplotlib>=1.4",
    "pillow",
]
docs_req = [
    "sphinx>=1.8",
    "numpydoc>=0.6",
    "nengo_sphinx_theme>=1.2.0",
    "nbsphinx>=0.8.8",
    "tensorflow",
    "jupyter",
]
optional_req = [
    "nengo-dl",
    "keras",
    "tensorflow",
    "numba>=0.43.1",
    "scipy",
]
tests_req = [
    "jupyter",
    "pytest>=6.0",
    "pytest-allclose",
    "pytest-plt",
    "pytest-rng",
]

setup(
    name="nengo-extras",
    version=version,
    author="Applied Brain Research",
    author_email="info@appliedbrainresearch.com",
    packages=find_packages(),
    url="https://www.nengo.ai/nengo-extras",
    include_package_data=False,
    license="Free for non-commercial use",
    description="Lesser used features for the Nengo neural simulator",
    long_description=read("README.rst", "CHANGES.rst"),
    zip_safe=False,
    install_requires=install_req,
    extras_require={
        "all": docs_req + optional_req + tests_req,
        "docs": docs_req,
        "optional": optional_req,
        "tests": tests_req,
    },
    python_requires=">=3.6",
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import nengo
import nengo.utils.numpy as npext
import numpy as np
from nengo.exceptions import ValidationError
from nengo.params import NumberParam

from nengo_extras import reqs

if reqs.HAS_SCIPY:
    import scipy.interpolate

if reqs.HAS_NUMBA:
    from numba import njit
    from numba.extending import overload

else:

    def njit(f):
        return f


def softplus(x, sigma=1.0):
    x = np.asarray(x)
    y = x / sigma
    z = np.array(x)
    z[y < 34.0] = sigma * np.log1p(np.exp(y[y < 34.0]))
    return z
    # ^ 34.0 gives exact answer in 32 or 64 bit but doesn't overflow in 32 bit


def lif_j(j, tau_ref, tau_rc, amplitude=1.0):
    return amplitude / (tau_ref + tau_rc * np.log1p(1.0 / j))


class SoftLIFRate(nengo.neurons.LIFRate):
    """LIF neuron with smoothing around the firing threshold.

    This is a rate version of the LIF neuron whose tuning curve has a
    continuous first derivative, due to the smoothing around the firing
    threshold. It can be used as a substitute for LIF neurons in deep networks
    during training, and then replaced with LIF neurons when running
    the network [1]_.

    Parameters
    ----------
    sigma : float
        Amount of smoothing around the firing threshold. Larger values mean
        more smoothing.
    amplitude : float
        Scaling factor on the output. If 1 (default), output rates correspond
        to LIF neuron model rates.
    tau_rc : float
        Membrane RC time constant, in seconds. Affects how quickly the membrane
        voltage decays to zero in the absence of input (larger = slower decay).
    tau_ref : float
        Absolute refractory period, in seconds. This is how long the
        membrane voltage is held at zero after a spike.

    References
    ----------
    .. [1] E. Hunsberger & C. Eliasmith (2015). Spiking Deep Networks with
       LIF Neurons. arXiv Preprint, 1510. https://export.arxiv.org/abs/1510.08829
    """

    sigma = NumberParam("sigma", low=0, low_open=True)

    def __init__(self, sigma=1.0, **lif_args):
        super().__init__(**lif_args)
        self.sigma = sigma  # smoothing around the threshold

    @property
    def _argreprs(self):
        args = super()._argreprs
        if self.sigma != 1.0:
            args.append("sigma=%s" % self.sigma)
        return args

    def rates(self, x, gain, bias):
        J = self.current(x, gain, bias)
        out = np.zeros_like(J)
        SoftLIFRate.step_math(self, dt=1, J=J, output=out)
        return out

    def step_math(self, dt, J, output):
        """Compute rates in Hz for input current (incl. bias)"""
        j = softplus(J - 1, sigma=self.sigma)
        output[:] = 0  # faster than output[j <= 0] = 0
        output[j > 0] = lif_j(
            j[j > 0], self.tau_ref, self.tau_rc, amplitude=self.amplitude
        )

    def derivative(self, x, gain, bias):
        y = gain * x + bias - 1
        j = softplus(y, sigma=self.sigma)
        yy = y[j > 0]
        jj = j[j > 0]
        vv = lif_j(jj, self.tau_ref, self.tau_rc, amplitude=self.amplitude)

        d = np.zeros_like(j)
        d[j > 0] = (gain * self.tau_rc * vv * vv) / (
            self.amplitude * jj * (jj + 1) * (1 + np.exp(-yy / self.sigma))
        )
        return d


class FastLIF(nengo.neurons.LIF):
    """Faster version of the leaky integrate-and-fire (LIF) neuron model.

    This neuron model is faster than ``LIF`` but does not produce the ideal
    firing rate for larger ``dt`` due to linearization of the tuning curves.
    """

    def step_math(self, dt, J, spiked, voltage, refractory_time):

        # update voltage using accurate exponential integration scheme
        dV = -np.expm1(-dt / self.tau_rc) * (J - voltage)
        voltage += dV
        voltage[voltage < self.min_voltage] = self.min_voltage

        # update refractory period assuming no spikes for now
        refractory_time -= dt

        # set voltages of neurons still in their refractory period to 0
        # and linearly reduce voltage when partway out of ref. period
        voltage *= (1 - refractory_time / dt).clip(0, 1)

        # determine which neurons spike (if v > 1 set spiked = 1/dt, else 0)
        spiked[:] = (voltage > 1) / dt

        # linearly approximate time since neuron crossed spike threshold
        overshoot = (voltage[spiked > 0] - 1) / dV[spiked > 0]
        spiketime = dt * (1 - overshoot)

        # set spiking neurons' voltages to zero, and ref. time to tau_ref
        voltage[spiked > 0] = 0
        refractory_time[spiked > 0] = self.tau_ref + spiketime


def spikes2events(t, spikes):
    """Return an event-based representation of spikes (i.e. spike times)"""
    spikes = npext.array(spikes, copy=False, min_dims=2)
    if spikes.ndim > 2:
        raise ValidationError(
            "Cannot handle %d-dimensional arrays" % spikes.ndim, attr="spikes"
        )
    if spikes.shape[-1] != len(t):
        raise ValidationError(
            "Last dimension of 'spikes' must equal 'len(t)'", attr="spikes"
        )

    # find nonzero elements (spikes) in each row, and translate to times
    return [t[spike != 0] for spike in spikes]


def _rates_isi_events(t, events, midpoint, interp):
    if not reqs.HAS_SCIPY:
        raise ImportError("`_rates_isi_events` requires `scipy`")

    if len(events) == 0:
        return np.zeros_like(t)

    isis = np.diff(events)

    rt = np.zeros(len(events) + (1 if midpoint else 2))
    rt[1:-1] = 0.5 * (events[:-1] + events[1:]) if midpoint else events
    rt[0], rt[-1] = t[0], t[-1]

    r = np.zeros_like(rt)
    r[1 : len(isis) + 1] = 1.0 / isis

    f = scipy.interpolate.interp1d(rt, r, kind=interp, copy=False)
    return f(t)


def rates_isi(t, spikes, midpoint=False, interp="zero"):
    """Estimate firing rates from spikes using ISIs.

    Parameters
    ----------
    t : (M,) array_like
        The times at which raw spike data (spikes) is defined.
    spikes : (M, N) array_like
        The raw spike data from N neurons.
    midpoint : bool, optional
        If true, place interpolation points at midpoints of ISIs. Otherwise,
        the points are placed at the beginning of ISIs.
    interp : string, optional
        Interpolation type, passed to `scipy.interpolate.interp1d` as the
        ``kind`` parameter.

    Returns
    -------
    rates : (M, N) array_like
        The estimated neuron firing rates.
    """
    spike_times = spikes2events(t, spikes.T)
    rates = np.zeros(spikes.shape)
    for i, st in enumerate(spike_times):
        rates[:, i] = _rates_isi_events(t, st, midpoint, interp)

    return rates


def lowpass_filter(x, tau, kind="expon"):
    nt = x.shape[-1]

    if kind == "expon":
        t = np.arange(0, 5 * tau)
        kern = np.exp(-t / tau) / tau
        delay = tau
    elif kind == "gauss":
        std = tau / 2.0
        t = np.arange(-4 * std, 4 * std)
        kern = np.exp(-0.5 * (t / std) ** 2) / np.sqrt(2 * np.pi * std ** 2)
        delay = 4 * std
    elif kind == "alpha":
        alpha = 1.0 / tau
        t = np.arange(0, 5 * tau)
        kern = alpha ** 2 * t * np.exp(-alpha * t)
        delay = tau
    else:
        raise ValidationError("Unrecognized filter kind '%s'" % kind, "kind")

    delay = int(np.round(delay))
    return np.array(
        [np.convolve(kern, xx, mode="full")[delay : nt + delay] for xx in x]
    )


def rates_kernel(t, spikes, kind="gauss", tau=0.04):
    """Estimate firing rates from spikes using a kernel.

    Parameters
    ----------
    t : (M,) array_like
        The times at which raw spike data (spikes) is defined.
    spikes : (M, N) array_like
        The raw spike data from N neurons.
    kind : str {'expon', 'gauss', 'expogauss', 'alpha'}, optional
        The type of kernel to use. 'expon' is an exponential kernel, 'gauss' is
        a Gaussian (normal) kernel, 'expogauss' is an exponential followed by
        a Gaussian, and 'alpha' is an alpha function kernel.
    tau : float
        The time constant for the kernel. The optimal value will depend on the
        firing rate of the neurons, with a longer tau preferred for lower
        firing rates. The default value of 0.04 works well across a wide range
        of firing rates.
    """
    spikes = spikes.T
    spikes = npext.array(spikes, copy=False, min_dims=2)
    if spikes.ndim > 2:
        raise ValidationError(
            "Cannot handle %d-dimensional arrays" % spikes.ndim, attr="spikes"
        )
    if spikes.shape[-1] != len(t):
        raise ValidationError(
            "Last dimension of 'spikes' must equal 'len(t)'", attr="spikes"
        )

    dt = t[1] - t[0]

    tau_i = tau / dt
    kind = kind.lower()
    if kind == "expogauss":
        rates = lowpass_filter(spikes, tau_i, kind="expon")
        rates = lowpass_filter(rates, tau_i / 4, kind="gauss")
    else:
        rates = lowpass_filter(spikes, tau_i, kind=kind)

    return rates.T


if reqs.HAS_NUMBA:

    @overload(np.clip)
    def np_clip(a, a_min, a_max):  # pragma: no cover
        """Numba-implementation of np.clip."""

        # Does not support `out` argument, optional arguments, nor a.clip
        # https://github.com/numba/numba/pull/3468
        def np_clip_impl(a, a_min, a_max):
            out = np.empty_like(a)
            for index, val in np.ndenumerate(a):
                if val < a_min:
                    out[index] = a_min
                elif val > a_max:
                    out[index] = a_max
                else:
                    out[index] = val
            return out

        return np_clip_impl


class NumbaLIF(nengo.LIF):
    """Numba-compiled version of the LIF model.

    Parameters
    ----------
    tau_rc : float
        Membrane RC time constant, in seconds. Affects how quickly the
        membrane voltage decays to zero in the absence of input
        (larger = slower decay).
    tau_ref : float
        Absolute refractory period, in seconds. This is how long the
        membrane voltage is held at zero after a spike.
    min_voltage : float
        Minimum value for the membrane voltage. If ``-np.inf``, the voltage
        is never clipped.
    amplitude : float
        Scaling factor on the neuron output. Corresponds to the relative
        amplitude of the output spikes of the neuron.
    """

    def __init__(self, *args, **kwargs):
        if not reqs.HAS_NUMBA:
            raise ImportError("`NumbaLIF` requires `numba`")
        super().__init__(*args, **kwargs)

    @staticmethod
    @njit
    def _lif_step_math(
        dt,
        J,
        spiked,
        voltage,
        refractory_time,
        tau_rc,
        tau_ref,
        min_voltage,
        amplitude,
    ):  # pragma: no cover
        # reduce all refractory times by dt
        refractory_time -= dt

        # compute effective dt for each neuron, based on remaining time.
        # note that refractory times that have completed midway into this
        # timestep will be given a partial timestep, and moreover these
        # will be subtracted to zero at the next timestep (or reset by a
        # spike)
        delta_t = np.clip(dt - refractory_time, 0, dt)

        # update voltage using discretized lowpass filter
        # since v(t) = v(0) + (J - v(0))*(1 - exp(-t/tau)) assuming
        # J is constant over the interval [t, t + dt)
        voltage -= (J - voltage) * np.expm1(-delta_t / tau_rc)

        # determine which neurons spiked (set them to 1/dt, else 0)
        spiked_mask = voltage > 1
        spiked[:] = spiked_mask * (amplitude / dt)

        # set v(0) = 1 and solve for t to compute the spike time
        t_spike = dt + tau_rc * np.log1p(
            -(voltage[spiked_mask] - 1) / (J[spiked_mask] - 1)
        )

        # set spiked voltages to zero, refractory times to tau_ref, and
        # rectify negative voltages to a floor of min_voltage
        voltage[voltage < min_voltage] = min_voltage
        voltage[spiked_mask] = 0
        refractory_time[spiked_mask] = tau_ref + t_spike

    def step_math(self, dt, J, spiked, voltage, refractory_time):
        self._lif_step_math(
            dt,
            J,
            spiked,
            voltage,
            refractory_time,
            self.tau_rc,
            self.tau_ref,
            self.min_voltage,
            self.amplitude,
        )
